#include <QtCore>
#include <bundleManager.h>
#include <bundle.h>
#include <readFile.h>
#include <nodeManager.h>
#include <bundleListFileIO.h>
#include <stdlib.h>

#define BUNDLEDESTUPDATE 1000
#define TTLTIMER 1000

BundleManager::BundleManager(int id)
{
	ourId=id;
	ReadFile conf;
	memoryStorageSize=conf.getMemoryStorageSize();
	fileStorageSize=conf.getStorageSize();
	fileOption=conf.getUseFileBundles();
	TTLOption=conf.getUseTTL();
	TTLValue=conf.getTTL();
	smartOffer=conf.getSmartOffer();
	if(TTLOption==1)
	{
		TTLTimer = new QTimer();
		TTLTimer->start(TTLTIMER);
		connect(TTLTimer, SIGNAL(timeout()), this, SLOT(checkTTL()));
	}
	routing=conf.getRouting();
	int WRITETIMER =conf.getWriteToFileTimer();
	storagePath=conf.getStoragePath();
	dir.setPath(storagePath);
	//Create the dir if it does not exist
	if(!dir.exists())
		dir.mkpath(dir.path());
	QString msgDirPath=conf.getMsgPath();
	msgDir.setPath(msgDirPath);
	option=conf.getUseFileBundles();
	logOption=conf.getLogOption();
	if(fileOption==1)
	{
		readFromFile();
		writeToFileTimer = new QTimer();
		connect(writeToFileTimer, SIGNAL(timeout()), this, SLOT(saveToFile()));
		writeToFileTimer->start(WRITETIMER);
		emit sendDebugBundleList(bundleList);
	}
	else
	{
		int seed = 10000;		/* choose a seed value */
		srand(seed);		/*initialize random number generator*/
		seq = (int)(rand()*1000000);
	}


	timer = new QTimer();
	showFirstTime=0;
	timer->start(5000);
	connect(timer, SIGNAL(timeout()), this, SLOT(emitDebugList()));
	if(logOption==1)
	{
		log = new MessageFile;
		log->set("transfer.log", 1000000, 5);
	}
	QString fileName = "bundleHash.txt";
	hashFile = new BundleListFileIO(&transferredBundleHash,
													 dir,
													 fileName);
	//get all bundle->node mappings from file

	int res; 
	if(conf.getUseFileHash()==1)
		res= hashFile->readAll();
	if(logOption==1)
	{
	    QString logString;
	    logString.append("Imported to hash from file, result(0=ok): ");	
	    logString.append(QString("%1").arg((int)res));
	    log->addLog(0,logString);
	}
}

void BundleManager::emitDebugList()
{
	if(showFirstTime==0)
	{
		emit sendDebugBundleList(bundleList);
		showFirstTime=1;
	}

}
void BundleManager::saveToFile()
{
	QFile file(dir.filePath("bundles.txt"));
	if(file.open(QIODevice::WriteOnly |QIODevice::Truncate| QIODevice::Text))
	{
	    QDateTime realTime;
	    realTime=realTime.currentDateTime ();
	    QString time=realTime.toString("MM/dd/yyyy hh:mm:ss");
	    time.append("\n");
	    file.write(time.toAscii ());
	    QString seqString;
	    seqString.append(QString("%1").arg((int)seq));
	    seqString.append("\n");
	    file.write(seqString.toAscii ());
	    for (ushort i = 0; i < bundleList.size(); ++i) 
	    {
		QString bundleString;
		bundleString.append(QString("%1").arg((int)bundleList.at(i).id));
		bundleString.append("\t");
		bundleString.append(QString("%1").arg((int)bundleList.at(i).sourceId));
		bundleString.append("\t");
		bundleString.append(QString("%1").arg((int)bundleList.at(i).destinationId));
		bundleString.append("\t");
		bundleString.append(bundleList.at(i).sourceString);
		bundleString.append("\t");
		bundleString.append(bundleList.at(i).destinationString);
		bundleString.append("\t");
		QDateTime Dtime = bundleList.at(i).timeStamp;
		QString time=Dtime.toString("MM/dd/yyyy hh:mm:ss");
		bundleString.append(time);
		bundleString.append("\t");
		bundleString.append(QString("%1").arg((int)bundleList.at(i).options));
		bundleString.append("\t");
		bundleString.append(QString("%1").arg((int)bundleList.at(i).dataLength));
		bundleString.append("\n");
		file.write(bundleString.toAscii ());
	    }
	    file.close();
	}
}

void BundleManager::readFromFile()
{
	QFile file(dir.filePath("bundles.txt"));
	if (file.open(QIODevice::ReadOnly|QIODevice::Text))
	{
		QString firstLine;
		QString seqLine;
		QString line;
		//First we read date and time
		firstLine=file.readLine();
		//Then we read bundle sequence number
		seqLine=file.readLine();
		seq=seqLine.toInt(0,10);
		//Then we add nodes
		line=file.readLine();
		while(line.size()>0)
		{
			//Parsing node
			int pos;
			QString temp;
			Bundle newBundle;
			//Bundle ID
			pos=line.indexOf("\t",0);
			temp=line.mid(0,pos);
			newBundle.id=temp.toInt(0,10);
			line.remove(0,pos+1);
			//Bundle source ID
			pos=line.indexOf("\t",0);
			temp=line.mid(0,pos);
			newBundle.sourceId=temp.toInt(0,10);
			line.remove(0,pos+1);
			//Bundle destination ID
			pos=line.indexOf("\t",0);
			temp=line.mid(0,pos);
			newBundle.destinationId=temp.toInt(0,10);
			line.remove(0,pos+1);
			//Bundle source string
			pos=line.indexOf("\t",0);
			temp=line.mid(0,pos);
			newBundle.sourceString=temp.toAscii();
			line.remove(0,pos+1);
			//Bundle destination string
			pos=line.indexOf("\t",0);
			temp=line.mid(0,pos);
			newBundle.destinationString=temp.toAscii();
			line.remove(0,pos+1);
			//Bundle timestamp
			pos=line.indexOf("\t",0);
			temp=line.mid(0,pos);
			QDateTime time;
			time=time.fromString(temp,"MM/dd/yyyy hh:mm:ss");
			newBundle.timeStamp=time;
			line.remove(0,pos+1);
			//Bundle options
			pos=line.indexOf("\t",0);
			temp=line.mid(0,pos);
			newBundle.options=temp.toInt(0,10);
			line.remove(0,pos+1);
			//Bundle datalength
			pos=line.indexOf("\t",0);
			temp=line.mid(0,pos);
			newBundle.dataLength=temp.toInt(0,10);
			line.remove(0,pos+1);
			//Adding Node
			QString bundlefilename;
			bundlefilename=QString("%1").arg((int)newBundle.id);
			QFile bundleFile(dir.filePath(bundlefilename));
			if (bundleFile.open(QIODevice::ReadOnly))
			{
			    int size=bundleFile.size();
			    bundleFile.close();
				if(bundleFile.size()>0)
				{
					bundleList.append(newBundle);
				}
			}


			line=file.readLine();
		}
		file.close();
	}
	//DEBUG
	emit sendDebugBundleList(bundleList);
	emit sendMemoryStorage(getMemoryStorageSize());
	emit sendFileStorage(getFileStorageSize());

}

QList<Bundle> BundleManager::getBundleList()
{	
	return bundleList;
}

int BundleManager::getMemoryStorageSize()
{	
	int size=0;
	for (ushort i = 0; i < bundleList.size(); ++i) 
	{
		size+=getMemoryBundleSize(bundleList.at(i));
	}
	return size;
}

int BundleManager::getFileStorageSize()
{	
	int size=0;
	for (ushort i = 0; i < bundleList.size(); ++i) 
	{
		size+=getFileBundleSize(bundleList.at(i));
	}
	return size;
}

int BundleManager::getMemoryBundleSize(Bundle bundle)
{	
	int size=0;
	size+=sizeof(Bundle);
	size+=bundle.data.size();
	size+=bundle.destinationString.size();
	size+=bundle.sourceString.size();
	return size;
}

int BundleManager::getFileBundleSize(Bundle bundle)
{	
	if(fileOption==1)
	{
		QString filename;
		filename=QString("%1").arg((int)bundle.id);
		QFile file(dir.filePath(filename));
		return file.size();
	}
	return 0;
}


QList<Bundle> BundleManager::getBundleOffer(QList<Node> ourList,QList<Node> encoutnerList,int encounterId)
{
	QList<Bundle> newBundleList;
//	return bundleList;

	switch(routing)
	{
		case 0:	//EPIDEMIC ROUTING - I like it.. ;)	
			//We have to check for each bundle
			for (ushort i = 0; i < bundleList.size(); ++i) 
			{
				//We have to exclude our own bundles
				if(bundleList.at(i).destinationId!=ourId)
				{		
					if(smartOffer==1)
					{
					    // only offer the bundle if it was not already sent to the 
					    // encountered node.
					    QList<int> l = transferredBundleHash.values(bundleList.at(i).id);
					    if(!l.contains(encounterId))
					    {
							Bundle tempBundle;
							tempBundle=bundleList.at(i);
							newBundleList.append(tempBundle);
							if(logOption==1)
							{
							    QString logString;
		    					logString.append("Bundle: ");	
							    logString.append(QString("%1").arg((int)tempBundle.id));
		    					logString.append(" with encounterId: ");	
							    logString.append(QString("%1").arg((int)encounterId));
		    					logString.append(" not in hash, adding it to offer.");	
		    					log->addLog(0,logString);
							}
						}else
						{
							if(logOption==1)
							{
							    QString logString;
		    					logString.append("Bundle: ");	
							    logString.append(QString("%1").arg((int)bundleList.at(i).id));
		    					logString.append(" with encounterId: ");	
							    logString.append(QString("%1").arg((int)encounterId));
		    					logString.append(" is in hash, not adding it to offer.");	
		    					log->addLog(0,logString);
							}
						}
					}
					else // no smartOffer
					{
					    Bundle tempBundle;
					    tempBundle=bundleList.at(i);
					    newBundleList.append(tempBundle);	
    				}
					
				}
			}
			return newBundleList;
				return bundleList; break;	//this does nothing, probably should be removed
		case 1: //PROPHET ROUTING: GRTR  P_(B,D)>P_(A,D) - A bit more advanced
		//We have to check for each bundle
			for (ushort i = 0; i < bundleList.size(); ++i) 
			{
				//We have to exclude our own bundles
				if(bundleList.at(i).destinationId!=ourId)
				{		
					float ourProbability = -1.0;
					float encounterProbability = -1.0;
					int bundleDestination = bundleList.at(i).destinationId;
					//First we get our probability
					for (ushort n = 0; n < ourList.size(); ++n) 
					{
						int nodeId=ourList.at(n).nodeId;
						if(nodeId==bundleDestination)
							ourProbability = ourList.at(n).probability;
						}
					//Then we get encouter probability
					for (ushort n = 0; n < encoutnerList.size(); ++n) 
					{
						int nodeId=encoutnerList.at(n).nodeId;
						if(encoutnerList.at(n).nodeId==bundleDestination)
							encounterProbability = encoutnerList.at(n).probability;
					}
					//GTRT and Epidemic for ACKS
					if((((encounterProbability>ourProbability)||(bundleDestination==encounterId))&&(bundleDestination!=0)))//||(((bundleList.at(i).options&0x02)==0x02)))
					{		
					    if(smartOffer==1)
			    	    {
							// only send the bundle if it was not already sent to the 
							// encountered node.
							QList<int> l = transferredBundleHash.values(bundleList.at(i).id);
							if(!l.contains(encounterId))
							{
							    Bundle tempBundle;
							    tempBundle=bundleList.at(i);
							    newBundleList.append(tempBundle);
								if(logOption==1)
								{
								    QString logString;
			    					logString.append("Bundle: ");	
								    logString.append(QString("%1").arg((int)tempBundle.id));
		    						logString.append(" with encounterId: ");	
								    logString.append(QString("%1").arg((int)encounterId));
		    						logString.append(" not in hash, adding it to offer.");	
		    						log->addLog(0,logString);
								}
    						}else
							{
								if(logOption==1)
								{
							    	QString logString;
		    						logString.append("Bundle: ");	
								    logString.append(QString("%1").arg((int)bundleList.at(i).id));
		    						logString.append(" with encounterId: ");	
								    logString.append(QString("%1").arg((int)encounterId));
		    						logString.append(" is in hash, not adding it to offer.");	
		    						log->addLog(0,logString);
								}
							}
					    }
					    else	//no smartOffer
					    {
					        Bundle tempBundle;
					        tempBundle=bundleList.at(i);
					        newBundleList.append(tempBundle);
    					    }
					}
				}
			}
			return newBundleList;
			break;
	}
	return bundleList;
}

void BundleManager::updateBundlesDestinations(NodeManager* nodeManager)
{
	for (ushort i = 0; i < bundleList.size(); ++i) 
	{
		Bundle tempBundle;
		if(bundleList.at(i).destinationId==0)
		{
			QString source;
			QString destination;
			QString temp;
			int pos;
			tempBundle=bundleList.at(i);
/*			temp=tempBundle.sourceString;
			pos=temp.indexOf("://",0);
			temp.remove(0,pos+3);
			pos=temp.indexOf("/",0);
			source=temp.mid(0,pos);*/
			temp=tempBundle.destinationString;
			pos=temp.indexOf("://",0);
			temp.remove(0,pos+3);
			pos=temp.indexOf("/",0);
			destination=temp.mid(0,pos);
			tempBundle.destinationId=nodeManager->getId(destination);
			bundleList.replace(i,tempBundle);
			emit sendDebugBundleList(bundleList);
		}
	} 

}

QList<Bundle> BundleManager::prepareBundels(QList<Bundle> receivedList, int senderId)
{
	//Add any acked bundles to the hash with bundle->node mappings, then remove 
	//it from the received list.
	for(int i=0; i<receivedList.size(); i++)
	{
		Bundle b = receivedList.at(i);
		if(b.isAck())
		{
			//Only add the sender to the hash if not alrady in it
			QList<int> l = transferredBundleHash.values(b.id);
			if(!l.contains(senderId)){
				transferredBundleHash.insertMulti(b.id,senderId);
				if(logOption==1){
					QString logString;
					logString.append("Received smart-offer ack for Bundle: ");	
					logString.append(QString("%1").arg((int)b.id));
					logString.append(", adding it to hash. Sender: ");	
					logString.append(QString("%1").arg((int)senderId));
					log->addLog(0,logString);
				}
			}
			receivedList.removeAt(i);
		}
	}
	
	QList<Bundle> newBundleList;
	//We have to find and add all the bundles with data to the list
	for (ushort i = 0; i < bundleList.size(); ++i) 
	{
		for (ushort j = 0; j < receivedList.size(); ++j) 
		{
			if(bundleList.at(i).id==receivedList.at(j).id)
				newBundleList.append(bundleList.at(i));
		}	
	}
	return newBundleList; 
}


QList<Bundle> BundleManager::prepareBundelRequest(QList<Bundle> receivedList)
{
	//We take only bundles that we don't have
	//Currently no memory restrictions!!!
	for (ushort i = 0; i < bundleList.size(); ++i) 
	{
		for (ushort j = 0; j < receivedList.size(); ++j) 
		{
			if((bundleList.at(i).id==receivedList.at(j).id))//||(bundleList.at(i).options==receivedList.at(j).id))
			{
				receivedList.removeAt(j);
				j=0;
			}
		}	
	}

	return receivedList; //We have to add some memory restrictions
}



void BundleManager::addBundleList(QList<Bundle> receivedList)
{
	for (ushort n = 0; n < receivedList.size(); ++n)
	{
		Bundle newBundle;
		newBundle=receivedList.at(n);
		addBundle(newBundle);
	}	
}


//This is internal slot for adding bundles to the list
void BundleManager::storeBundle(Bundle newBundle)
{

	//we obtain space for bundle
	//FIFO 
	while( 
		 (((getMemoryStorageSize()+getMemoryBundleSize(newBundle))>memoryStorageSize)||
		 ((getFileStorageSize()+getFileBundleSize(newBundle))>fileStorageSize))&&
		 (bundleList.size()>0)
		 )
	{
		if(fileOption==1)
		{
			Bundle firstBundle;
			firstBundle=bundleList.first();
			QString filename;
			filename=QString("%1").arg((int)firstBundle.id);
			QFile file(dir.filePath(filename));
			file.remove();
		}
		bundleList.removeFirst();
	}
	//Then we add it to the list
	if(fileOption==1)
	{
		QString filename;
		filename=QString("%1").arg((int)newBundle.id);
		QFile file(dir.filePath(filename));
		if (file.open(QIODevice::WriteOnly))
		{
			file.write(newBundle.data);
			newBundle.data.clear();
			file.close();
		}
	}
	bundleList.append(newBundle);

		
	//DEBUG
	emit sendDebugBundleList(bundleList);
	emit sendMemoryStorage(getMemoryStorageSize());
	emit sendFileStorage(getFileStorageSize());



}

//This is external slot for receivng bundles
void BundleManager::addBundle(Bundle newBundle)
{
	int isInTheList=0;
 	Bundle tempBundle=newBundle;
	//Checking that we don't have the same bundle
	//or that wedidn't get any ACK for this bundle sofar
	for (ushort n = 0; n < bundleList.size(); ++n)
	{
		if(tempBundle.id==bundleList.at(n).id)
			isInTheList++;
		//Is there allready an ack for this bundle
		if((bundleList.at(n).options&0x02)==0x02)
		{		
			QByteArray data;
			if(fileOption==1)
			{
				QString filename;
				filename=QString("%1").arg((int)bundleList.at(n).id);
				QFile file(dir.filePath(filename));
				if(file.exists()==1)
				{
					if (file.open(QIODevice::ReadOnly))
					{
						data=file.readAll();
			    			file.close();
					}
				}
			}
			else
			{
				data = tempBundle.data;
			}
			if(tempBundle.id==data.toInt(0,10))   //Checking the bundle ACK bit
				isInTheList++;
		}
	}	
	if(isInTheList==0)
	{
		//We add a new bundle record to log file
 		if(logOption==1)
		{
		    QString logString;
		    logString.append("Bundle ID:");	
		    logString.append(QString("%1").arg((int)tempBundle.id));
		    logString.append(" SRC:");
		    logString.append(QString("%1").arg((int)tempBundle.sourceId));
		    logString.append(" ");
		    logString.append(tempBundle.sourceString);
		    logString.append(" DST:");
		    logString.append(QString("%1").arg((int)tempBundle.destinationId));
		    logString.append(" ");
		    logString.append(" Size:");
		    logString.append(QString("%1").arg((int)tempBundle.data.size()));
		    logString.append(tempBundle.destinationString);
		    log->addLog(0,logString);
		}
		//If this is a bundle endpoint we add it	
		if(tempBundle.destinationId==ourId)
		{
			//First we check if the Bundle is not ACK
			if((tempBundle.options&0x02)!=0x02)//Checking the bundle ACK bit
			{
				//We pass the bundle to Application layer
				//If this is a PRoPHET message...
				if((tempBundle.options&0x04)==0x04)//Checking the "PRoPHET" message bit
				{
					//We add a new msg
					QString filename;
					filename.append(tempBundle.timeStamp.toString("MM-dd_hh-mm"));
					filename.append("_ID_");
					filename.append(QString("%1").arg((int)tempBundle.id));
					filename.append("_FROM_");
					filename.append(tempBundle.sourceString);
					QFile file(msgDir.filePath(filename));
					if(file.open(QIODevice::WriteOnly|QIODevice::Truncate))
					{
						file.write(tempBundle.data);
						file.close();
					}
				}
				//Otherwise we pass it to the DTN
				else
				{
					QString logString;
					logString.append("Adding bundle to DTN  ID:");	
					logString.append(QString("%1").arg((int)tempBundle.id));
					logString.append(" SRC:");
					logString.append(QString("%1").arg((int)tempBundle.sourceId));
					logString.append(" ");
					logString.append(tempBundle.sourceString);
					logString.append(" DST:");
					logString.append(QString("%1").arg((int)tempBundle.destinationId));
					logString.append(" ");
					logString.append(tempBundle.destinationString);
					logString.append(" Size:");
					logString.append(QString("%1").arg((int)tempBundle.data.size()));
				    log->addLog(0,logString);
					emit sendDeliveredBundle(tempBundle);
				}
				//We store the bundle to our list
				//Adding recived bundle to storage if we are the ENDPOINT!!!!!!!!!!!!!
				storeBundle(tempBundle);
				//Then we check if we need to ACK this bundle
				if((tempBundle.options&0x01)==0x01)
				{
					//We create ACK
					Bundle ACKBundle;
					//And put the ACKed bundle ID to data field
					QByteArray ACKData;
					ACKData.setNum(tempBundle.id);
					int ACKOption;
					ACKOption = tempBundle.options;
					//We clear the ACKOption bit and set the ACK bit 
					ACKOption=ACKOption|0x02;
					ACKOption=ACKOption&0xFFFE;
					ACKBundle.setContent(seq,tempBundle.destinationId,tempBundle.sourceId,ACKData,ACKOption,tempBundle.destinationString,tempBundle.sourceString);
					seq++;
					//and add it to the list
					storeBundle(ACKBundle);
				}
			}
			//This is the ACK Bundle for us
			else
			{
				//Adding recived bundle to storage
				storeBundle(tempBundle);
				//Here can we pass some ACK inforation for application layer
				qint32 bundleId=tempBundle.data.toInt(0,10);
				removeBundle(bundleId);
			}
		}
		//This is not bundle's endpoint
		else
		{
			//If this is ACK of one of bundle that we have we remove it 
			if((tempBundle.options&0x02)==0x02)//Checking the bundle ACK bit
			{
				qint32 bundleId=tempBundle.data.toInt(0,10);
				removeBundle(bundleId);
			}
			storeBundle(tempBundle);
		}

	}
}

bool BundleManager::removeBundle(int bundleID)
{
	//First we search for the bundle
	for (ushort i = 0; i < bundleList.size(); ++i) 
	{
		if(bundleID==bundleList.at(i).id)
		{
			//Then we dele it
			if(fileOption==1)
			{
				QString filename;
				filename=QString("%1").arg((int)bundleList.at(i).id);
				QFile file(dir.filePath(filename));
				if(file.exists()==1)
				{
					file.remove();
				}
			}
			bundleList.removeAt(i);			
			//DEBUG
			emit sendDebugBundleList(bundleList);
			emit sendMemoryStorage(getMemoryStorageSize());
			emit sendFileStorage(getFileStorageSize());
			return 1;
		}
	}
	return 0;

}


void BundleManager::checkTTL()
{
	if(TTLOption==1)
	{
		QDateTime currentTime;
		currentTime=currentTime.currentDateTime();
		int currentTimeInSeconds=currentTime.toTime_t();
		for (ushort i = 0; i < bundleList.size(); ++i) 
		{
			int bundleTime=bundleList.at(i).timeStamp.toTime_t();
			if((bundleList.at(i).timeStamp.toTime_t()+TTLValue)<currentTimeInSeconds)
			{
				//Then we delete it
				removeBundle(bundleList.at(i).id);
				i=0;
			}
		}
	}
}

void BundleManager::removeAllBundles()
{
	while(bundleList.size()>0)
	{
		removeBundle(bundleList.first().id);
	}
}

int BundleManager::getSeq()
{
	seq++;
	seq%=1000000;
	return seq;
}


