#include <QtCore>
#include <connection.h>
#include <tlv.h>
#include <messageFile.h>
#include <dataPacket.h>
#include <readFile.h>

#define NOW 100

//#define HELLO_TIMER 60000
//#define INITIATOR_TIMER 60000
//#define LISTENER_TIMER 60000
//#define HELLOT 10000
//#define BROADCAST 1000

//Connections states for GUI
#define STATE_SYN 1
#define STATE_SYNACK 2
#define STATE_ACK 3
#define STATE_DICT 4
#define STATE_RIB 5
#define STATE_BOFF 6
#define STATE_BREQ 7
#define STATE_BUNDLE 8
#define STATE_ERROR 9

//Connection states
#define WAIT_NB 0
#define HELLO 1
#define INFO_EXCH 2
#define WAIT_INFO 2


//Hello procedure states
#define SENDSYN 1
#define WAITSYNACK 2
#define WAITACK 4
#define ESTABLISHED 3
#define MASTER 4
#define SLAVE 5

//Initiator states
#define SENDDICTIONARY 1
#define SENDRIB 2
#define SENDBOFFER 4
#define WAITBREQ 5
#define TIMEOUT 16
#define SENDBUNDLE 7
#define WAITFORBREQUEST 8
#define RESETLINK 19
#define IDLESTATE 110

//Listener states
#define WAITFORDICTIONARY 1
#define WAITFORRIB 2
#define WAITFORBOFFER 3
#define SENDBREQUEST 4
#define WAITFORBUNDLES 5



Connection::Connection(int connectId,int nodeId,QString nodeName,NodeManager* nodeManager,BundleManager* bundleManager,QObject *parent)

{
	ReadFile init;
	HELLO_TIMER=init.getHelloTimer();
	INITIATOR_TIMER=init.getInitiatorTimer();
	LISTENER_TIMER=init.getListenerTimer();
	HELLOT=init.getHello();
	BROADCAST=init.getBroadcastTimer();
	continiusUpdate=init.getContiniusUpdate();
	update=0;
	updateRIB=0;
	encounterNodeId=0;
	//State variables
	connectionState=HELLO;
	helloState=SENDSYN;
	initiatorState=IDLESTATE;
	listenerState=IDLESTATE;
	//Define our node manager
	connectionNodeManager=nodeManager;
	//Define our bundle manager
	connectionBundleManager=bundleManager;
	//Set up unique connection ID
//	connectionId=rand();
//	connectionId=connectionId%1000;
	connectionId=connectId;
	//Set up connectionNodeName
	connectionNodeName=nodeName;
	//Set up connectionNodeName
	connectionNodeId=nodeId;
	//Create and connect Hello Timer
    helloTimer = new QTimer(this);
    connect(helloTimer, SIGNAL(timeout()), this, SLOT(helloProcedure()));
    helloTimer->start(rand()%HELLO);

	//Create and connect Initiator Timer
    initiatorTimer = new QTimer(this);
    connect(initiatorTimer, SIGNAL(timeout()), this, SLOT(initiatorProcedure()));
    //initiatorTimer->start(INITIATOR_TIMER);

	//Create and connect Listener Timer
    listenerTimer = new QTimer(this);
    connect(listenerTimer, SIGNAL(timeout()), this, SLOT(listenerProcedure()));
    //listenerTimer->start(LISTENER_TIMER);




	//Create and connect TLV translator
	TLVTranslator = new TLV;

	//Main external datastreams
	connect(TLVTranslator,SIGNAL(sendDatagram(QByteArray)),this,SLOT(forwardSlotDatagram(QByteArray)));
	connect(this,SIGNAL(forwardSignalDatagram(QByteArray)),TLVTranslator,SLOT(receiveDatagram(QByteArray)));

	//Hello procedure connections
    connect(this, SIGNAL(sendHello(Hello)),TLVTranslator,SLOT(receiveHello(Hello)));
    connect(TLVTranslator, SIGNAL(sendHello(Hello)),this,SLOT(receiveHello(Hello)));

	//Initiators procedure connections
    connect(this, SIGNAL(sendDictionary(QList<Node>)),TLVTranslator,SLOT(createDictionary(QList<Node>)));
    connect(this, SIGNAL(sendRIB(QList<Node>)),TLVTranslator,SLOT(createRIB(QList<Node>)));
    connect(this, SIGNAL(sendBundleOffer(QList<Bundle>)),TLVTranslator,SLOT(createBundleOffer(QList<Bundle>)));
    connect(TLVTranslator, SIGNAL(sendBundleRequest(QList<Bundle>)),this,SLOT(receiveBundleRequest(QList<Bundle>)));
	connect(this, SIGNAL(sendBundles(QList<Bundle>)),TLVTranslator,SLOT(createBundleData(QList<Bundle>)));


	//Listener procedure connections
    connect(TLVTranslator, SIGNAL(sendDictionary(QList<Node>)),this,SLOT(receiveDictionary(QList<Node>)));
    connect(TLVTranslator, SIGNAL(sendRIB(QList<Node>)),this,SLOT(receiveRIB(QList<Node>)));
    connect(TLVTranslator, SIGNAL(sendBundleOffer(QList<Bundle>)),this,SLOT(receiveBundleOffer(QList<Bundle>)));
    connect(this, SIGNAL(sendBundleRequest(QList<Bundle>)),TLVTranslator,SLOT(createBundleRequest(QList<Bundle>)));
    connect(TLVTranslator, SIGNAL(sendBundleData(QList<Bundle>)),this,SLOT(receiveBundles(QList<Bundle>)));
	//GUI state
	emit newConnectionStatus(STATE_SYN);

}


Connection::~Connection(void)
{
	if(helloTimer!=NULL)
	{
		helloTimer->stop();
		delete helloTimer;
	}
		if(initiatorTimer!=NULL)
	{
		initiatorTimer->stop();
		delete initiatorTimer;
	}
			if(listenerTimer!=NULL)
	{
		listenerTimer->stop();
		delete listenerTimer;
	}	
	if(TLVTranslator!=NULL)
	{
		delete TLVTranslator; 
	} 

}


void Connection::helloProcedure()
{
	helloTimer->stop();
	//Prepare new Hello Object
	Hello newHello;
	switch(helloState)
	{
		case SENDSYN:	//log->addLog(connectionId,(QString)"Hello Procedure:SENDSYN");
						newHello.setFunction(SYN);
						newHello.setTimer(HELLO_TIMER);
						newHello.setSenderName(connectionNodeName);
						newHello.senderId=connectionNodeId;
						newHello.senderType=0;
						helloState=WAITSYNACK;
						emit sendHello(newHello);
						//GUI state
						emit newConnectionStatus(STATE_SYNACK);
			
		break;			
		default	:		//log->addLog(connectionId,(QString)"Hello Procedure:Default");
						helloState=SENDSYN;
						connectionState=HELLO;
						initiatorState=IDLESTATE;
						listenerState=IDLESTATE;
						listenerTimer->stop();
						initiatorTimer->stop();	
						//GUI state
						emit newConnectionStatus(STATE_ERROR);
	
		break;
		}	
		helloTimer->start(HELLO_TIMER);
}



void Connection::receiveHello(Hello receivedHello)
{
	int helloType=receivedHello.getFunction();
	helloTimer->stop();
	
	//Prepare new Hello object
	Hello newHello;
	
	switch(helloType)
	{
	case SYNACK:	//log->addLog(connectionId,(QString)"Receive Hello:SYNACK");
					//Resolving a ID form encountered node
					encounterNodeId=receivedHello.senderId;
					encounterNodeName=receivedHello.senderName;
					encounterNodeType=receivedHello.senderType;
					if(update==0)
					{
						connectionNodeManager->encounterNode(encounterNodeId,encounterNodeType,encounterNodeName);
						if(continiusUpdate==0)
							update=1;
					}

					//Setting our own ID
					newHello.setFunction(ACK);
					newHello.setTimer(HELLO_TIMER);
					newHello.setSenderName(connectionNodeName);
					newHello.senderId=connectionNodeId;
					newHello.senderType=0;
					emit sendHello(newHello);
					helloState=SLAVE;
					//Invoke Initiator and Listener
					helloTimer->stop();
					initiatorState=SENDDICTIONARY;
					listenerState=WAITFORDICTIONARY;
					initiatorTimer->start(NOW);
					listenerTimer->start(LISTENER_TIMER);
					///initiatorProcedure();
					//////////////////////////////
					//GUI state
					emit newConnectionStatus(STATE_DICT);
					
	break;
	case SYN:		//log->addLog(connectionId,(QString)"Receive Hello:SYN");
					newHello.setFunction(SYNACK);
					newHello.setTimer(HELLO_TIMER);
					newHello.setSenderName(connectionNodeName);
					newHello.senderId=connectionNodeId;
					newHello.senderType=0;
					helloState=WAITACK;helloTimer->start(HELLO_TIMER);
					emit sendHello(newHello);
					//GUI state
					emit newConnectionStatus(STATE_SYNACK);

	break;
	case ACK:		//log->addLog(connectionId,(QString)"Receive Hello:ACK");
					//Resolving a ID form encountered node
					encounterNodeId=receivedHello.senderId;
					encounterNodeName=receivedHello.senderName;
					encounterNodeType=receivedHello.senderType;
					if(update==0)
					{
						connectionNodeManager->encounterNode(encounterNodeId,encounterNodeType,encounterNodeName);
						if(continiusUpdate==0)
							update=1;
					}
					helloState=MASTER;
					//Invoke Initiator and Listener
					helloTimer->stop();
					initiatorState=SENDDICTIONARY;
					listenerState=WAITFORDICTIONARY;
					initiatorTimer->start(NOW);
					listenerTimer->start(LISTENER_TIMER);
					//GUI state
					emit newConnectionStatus(STATE_DICT);
	break;
	case RSTACK:	//log->addLog(connectionId,(QString)"Receive Hello:RSTACK");
					helloState=SENDSYN;
					//GUI state
					emit newConnectionStatus(STATE_ERROR);
					connectionState=HELLO;helloTimer->start(HELLO_TIMER);
	break;
	default:		//log->addLog(connectionId,(QString)"Receive Hello:Default");
					helloState=SENDSYN;
					connectionState=HELLO;helloTimer->start(HELLO_TIMER);
					//GUI state
					emit newConnectionStatus(STATE_ERROR);
	break;
	}
	

}

void Connection::listenerProcedure()
{
	listenerTimer->stop();
	switch(initiatorState)
		{
		case WAITFORDICTIONARY://log->addLog(connectionId,(QString)"Listener:DICTIONARY timeout...");
							//we reset the link
							initiatorState=IDLESTATE;
							listenerState=IDLESTATE;
							connectionState=HELLO;
							helloState=SENDSYN;
							listenerTimer->stop();
							initiatorTimer->stop();
							helloTimer->start(HELLO_TIMER);
							//GUI state
							emit newConnectionStatus(STATE_ERROR);

							
			
		break;			
		case WAITFORRIB:	//log->addLog(connectionId,(QString)"Listener:RIB timeout...");
							//we reset the link
							initiatorState=IDLESTATE;
							listenerState=IDLESTATE;
							connectionState=HELLO;
							helloState=SENDSYN;
							listenerTimer->stop();
							initiatorTimer->stop();
							helloTimer->start(HELLO_TIMER);
							//GUI state
							emit newConnectionStatus(STATE_ERROR);
			
		break;
		case WAITFORBOFFER:	//log->addLog(connectionId,(QString)"Listener:BOFFER timeout...");
							//we reset the link
							initiatorState=IDLESTATE;
							listenerState=IDLESTATE;
							connectionState=HELLO;
							helloState=SENDSYN;
							listenerTimer->stop();
							initiatorTimer->stop();
							helloTimer->start(HELLO_TIMER);
							//GUI state
							emit newConnectionStatus(STATE_ERROR);

			
		break;
		case WAITFORBREQUEST://log->addLog(connectionId,(QString)"Listener:BREQ Timeout...");
							//we reset the link
							initiatorState=IDLESTATE;
							listenerState=IDLESTATE;
							connectionState=HELLO;
							helloState=SENDSYN;
							listenerTimer->stop();
							initiatorTimer->stop();
							helloTimer->start(HELLO_TIMER);
							//GUI state
							emit newConnectionStatus(STATE_ERROR);

			
		break;
		case WAITFORBUNDLES://log->addLog(connectionId,(QString)"Listener:BUNDLES Timeout...");
							//we reset the link
							initiatorState=IDLESTATE;
							listenerState=IDLESTATE;
							connectionState=HELLO;
							helloState=SENDSYN;
							listenerTimer->stop();
							initiatorTimer->stop();
							helloTimer->start(HELLO_TIMER);
							listenerState=RESETLINK;
							//GUI state
							emit newConnectionStatus(STATE_ERROR);

			
		break;
		case IDLESTATE:		//log->addLog(connectionId,(QString)"Listener:Idle state...");
							//we reset the link
							initiatorState=IDLESTATE;
							listenerState=IDLESTATE;
							connectionState=HELLO;
							helloState=SENDSYN;
							listenerTimer->stop();
							initiatorTimer->stop();
							helloTimer->start(HELLO_TIMER);
							//GUI state
							emit newConnectionStatus(STATE_ERROR);
														
							
		break;
		case RESETLINK:		//log->addLog(connectionId,(QString)"Listener:Link problem...");
							//we reset the link
							initiatorState=IDLESTATE;
							listenerState=IDLESTATE;
							connectionState=HELLO;
							helloState=SENDSYN;
							listenerTimer->stop();
							initiatorTimer->stop();
							helloTimer->start(HELLO_TIMER);
							//GUI state
							emit newConnectionStatus(STATE_ERROR);
			
		break;
		default	://log->addLog(connectionId,(QString)"Hm Listener Timeout default...");		
		break;
		}	
}


void Connection::initiatorProcedure()
{
	initiatorTimer->stop();
	QList<Node> nodeList;
	QList<Bundle> bundleList;
	switch(initiatorState)
		{
		case SENDDICTIONARY://log->addLog(connectionId,(QString)"Initiator:SENDDICTIONARY-RIB-BUNDLEOFFER");
							nodeList = connectionNodeManager->getNodeList();
							initiatorState=SENDRIB;
							emit sendDictionary(nodeList);
							nodeList = connectionNodeManager->getRIBNodeList();//We have to use only nodes with RIB
							initiatorState=SENDBOFFER;
							emit sendRIB(nodeList);
							bundleList = connectionBundleManager->getBundleOffer(connectionNodeManager->getNodeList(),lastReceivedRIB,encounterNodeId);
							emit sendBundleOffer(bundleList);
							initiatorState=WAITBREQ;
							initiatorTimer->start(INITIATOR_TIMER);
							//GUI state
							emit newConnectionStatus(STATE_BREQ);
		
		break;
		case WAITFORBREQUEST://log->addLog(connectionId,(QString)"Initiator:BREQ Timeout");
							//we reset the link
							initiatorState=IDLESTATE;
							listenerState=IDLESTATE;
							connectionState=HELLO;
							helloState=SENDSYN;
							listenerTimer->stop();
							initiatorTimer->stop();
							helloTimer->start(HELLO_TIMER);
							//GUI state
							emit newConnectionStatus(STATE_ERROR);
			
		break;
		case IDLESTATE:		//log->addLog(connectionId,(QString)"Initiator:Idle state...");
							//we reset the link
							initiatorState=IDLESTATE;
							listenerState=IDLESTATE;
							connectionState=HELLO;
							helloState=SENDSYN;
							listenerTimer->stop();
							initiatorTimer->stop();
							helloTimer->start(HELLO_TIMER);
							//GUI state
							emit newConnectionStatus(STATE_ERROR);

			
		break;
		case RESETLINK:		//log->addLog(connectionId,(QString)"Initiator:Link problem...");
							//we reset the link
							initiatorState=IDLESTATE;
							listenerState=IDLESTATE;
							connectionState=HELLO;
							helloState=SENDSYN;
							listenerTimer->stop();
							initiatorTimer->stop();
							helloTimer->start(HELLO_TIMER);
							//GUI state
							emit newConnectionStatus(STATE_ERROR);
			
		break;
		default	://log->addLog(connectionId,(QString)"Hm default timeout initiator...");		
		break;
		}	
 }


void Connection::receiveBundleOffer(QList<Bundle> receivedList) //LISTENER
{
	listenerTimer->stop();
	bundleRequestList=connectionBundleManager->prepareBundelRequest(receivedList);
	//log->addLog(connectionId,(QString)"Listener:Received Bundle Offer");
//	we send only one bundle request
	if(bundleRequestList.size()>0)
	{
		//log->addLog(connectionId,(QString)"Listener:Sending Full bundle request...");
		QList<Bundle> dummyList;
		dummyList.append(bundleRequestList.first());
		bundleRequestList.removeFirst();
		listenerState=WAITFORBUNDLES;
		listenerTimer->start(LISTENER_TIMER);
		emit sendBundleRequest(dummyList);
		//GUI state
		emit newConnectionStatus(STATE_BUNDLE);

	}
	else 
	{
		//log->addLog(connectionId,(QString)"Listener:Sending Empty bundle request...");
		if(initiatorState==IDLESTATE)
		{	
			//we reset the link
			initiatorState=IDLESTATE;
			listenerState=IDLESTATE;
			listenerTimer->stop();
			initiatorTimer->stop();

			if(helloState==MASTER)
			{
				helloState=SENDSYN;
				helloTimer->start(NOW);
				//log->addLog(connectionId,(QString)"Listener(receiveBOffer):We can restart connection...(MASTER)");
				//GUI state
				emit newConnectionStatus(STATE_SYN);

			}
			else
			{
				helloState=SENDSYN;
				helloTimer->start(HELLO_TIMER);
				//GUI state
				emit newConnectionStatus(STATE_SYN);

				//log->addLog(connectionId,(QString)"Listener(receiveBOffer):We can not restart connection(SLAVE)");

			}
		}
		else
		{
			//log->addLog(connectionId,(QString)"Listener(receiveBOffer):But we can not restart connection...");
			listenerState=IDLESTATE;
			listenerTimer->stop();
		}
		emit sendBundleRequest(bundleRequestList);
		
	}

}

void Connection::receiveBundles(QList<Bundle> receivedList) //LISTENER
{
	//log->addLog(connectionId,(QString)"Listener:Received Bundles...");
	listenerTimer->stop();
	connectionBundleManager->addBundleList(receivedList);
//	we send only one bundle request
	if(bundleRequestList.size()>0)
	{
		//log->addLog(connectionId,(QString)"Listener:Requesting Bundles...");
		QList<Bundle> dummyList;
		for(int i=0; i<receivedList.size(); i++)
		{
			Bundle b = receivedList.at(i);
			b.setAck();
			//should we reset the ack-bit afterwards?
			dummyList.prepend(b);
			//do some logging here
		}
		dummyList.append(bundleRequestList.first());
		bundleRequestList.removeFirst();
		listenerState=WAITFORBUNDLES;
		listenerTimer->start(LISTENER_TIMER);
		emit sendBundleRequest(dummyList);
		//GUI state
		emit newConnectionStatus(STATE_BUNDLE);

	}
	else //we finish comunication by sending empty bundle request
	{
		//log->addLog(connectionId,(QString)"Listener:No more needed Bundles...");
		if(initiatorState==IDLESTATE)
		{	
			//we reset the link
			initiatorState=IDLESTATE;
			listenerState=IDLESTATE;
			listenerTimer->stop();
			initiatorTimer->stop();

			if(helloState==MASTER)
			{
				helloState=SENDSYN;
				helloTimer->start(HELLOT);
				//log->addLog(connectionId,(QString)"Listener(receiveBundles):We can restart connection...(MASTER)");
				//GUI state
				emit newConnectionStatus(STATE_SYN);

			}
			else
			{
				//helloState=SENDSYN;
				//helloTimer->start(HELLO_TIMER);
				//log->addLog(connectionId,(QString)"Listener(receiveBundles):We can not restart connection(SLAVE)");

			}
		}
		else
		{
			//log->addLog(connectionId,(QString)"Listener(receiveBundles):But we can not restart connection...");
			listenerState=IDLESTATE;
			listenerTimer->stop();
		
		}
		//append the received bundles (as acks) to the next request.
		//TODO: Should this be moved to the bundle manager???
		for(int i=0; i<receivedList.size(); i++)
		{
			Bundle b = receivedList.at(i);
			b.setAck();
			//should we reset the ack-bit afterwards?
			bundleRequestList.prepend(b);
			//do some logging here
		}
		emit sendBundleRequest(bundleRequestList);
		
	}
		
}

void Connection::receiveBundleRequest(QList<Bundle> receivedList) //INITIATOR
{
	initiatorTimer->stop();
	QList<Bundle> bundleList;
	if(receivedList.size()>0)
	{
		bundleList=connectionBundleManager->prepareBundels(receivedList, encounterNodeId);
		initiatorState=WAITBREQ;
		//log->addLog(connectionId,(QString)"Initiator:Sending bundles");
		initiatorTimer->start(INITIATOR_TIMER);
		emit sendBundles(bundleList);
		//GUI state
		emit newConnectionStatus(STATE_BREQ);

	}
	else
	{
		//log->addLog(connectionId,(QString)"Initiator:No more requested bundles");
		if(listenerState==IDLESTATE)
		{	
			//we reset the link
			if(helloState==MASTER)
			{
				helloState=SENDSYN;
				helloTimer->start(HELLOT);
				//log->addLog(connectionId,(QString)"Listener(receiveBundles):We can restart connection...");
				//GUI state
				emit newConnectionStatus(STATE_SYN);

			}
			else
			{
				helloState=SENDSYN;
				helloTimer->start(HELLO_TIMER);
				//log->addLog(connectionId,(QString)"Listener(receiveBundles):We can not restart connection(SLAVE)");
				//GUI state
				emit newConnectionStatus(STATE_SYN);

			}
		}
		else
		{
			//log->addLog(connectionId,(QString)"Initiator:But we can not restart connection...");
			initiatorState=IDLESTATE;
			initiatorTimer->stop();
		}
	}
	
}


void Connection::receiveDictionary(QList<Node> receivedList)
{
	listenerTimer->stop();
	connectionNodeManager->addDictionary(connectionNodeId,receivedList);
	listenerState=WAITFORRIB;
	//log->addLog(connectionId,(QString)"Listener:Received dictionary");
	listenerTimer->start(LISTENER_TIMER);
	//GUI state
	emit newConnectionStatus(STATE_RIB);

}

void Connection::receiveRIB(QList<Node> receivedList)
{
	listenerTimer->stop();
	if(updateRIB==0)
	{		
		connectionNodeManager->addRIB(receivedList,encounterNodeId);
		if(continiusUpdate==0)
			updateRIB=1;
	}		
	listenerState=WAITFORBOFFER;
	//Needed for Routing
	lastReceivedRIB=receivedList;
	//log->addLog(connectionId,(QString)"Listener:Received RIB");
	listenerTimer->start(LISTENER_TIMER);
	//GUI state
	emit newConnectionStatus(STATE_BOFF);

}



void Connection::receiveDatagram(DataPacket dataPacket)
{
	////log->addLog(connectionId,(QString)"Received datagram...");
	emit forwardSignalDatagram(dataPacket.data);
}


void Connection::forwardSlotDatagram(QByteArray datagram)
{
	DataPacket dataPacket;
	////log->addLog(connectionId,(QString)"Sended datagram...");
	dataPacket.data = datagram;
	dataPacket.id = connectionId;
	emit sendDataPacket(dataPacket);
}
