#include <debugWidget.h>
#include <QtGui>
#include <QtCore>
#include <readFile.h>


#define DTN 500

DebugWidget::DebugWidget(NodeManager *nodeMng,BundleManager *bundleMng,QWidget *parent)
		: QWidget(parent)
{
#ifdef PDAGUI
	setFixedSize(240,320);
#else
	setFixedSize(1000,700);
#endif //PDAGUI
	ReadFile conf;
	QString mainText = "Prophet Window - ";
	mainText.append(conf.getNodeName());
	setWindowTitle(mainText);
	bundleManager=bundleMng;

#ifdef PDAGUI
	//Connection list tab
	connectionWidget = new ConnectionWidget(this);
	
	//Read Messages Tab
	messageRead = new MsgReadWidget(nodeMng,bundleMng);

	//Configuration Tab
	configWidget = new ConfigWidget(this);
#endif //PDAGUI

	//Message Widget Tag
	message = new MsgWidget(nodeMng,bundleMng);
	connect(message, SIGNAL(sendBundle(Bundle)), this, SLOT(forwardAppBundle(Bundle)));

	//Node Table Tag
	nodeTable = new QTableWidget(40, 4, this);
    QStringList horizontalNodeLabels;
    horizontalNodeLabels << "ID" << "Name" << "Probability" << "RIB Flag";
	nodeTable->setHorizontalHeaderLabels(horizontalNodeLabels); 

	//Info tag
	infoWidget = new InfoWidget(nodeMng);
	connect(this, SIGNAL(traffic(int)), infoWidget, SLOT(getTraffic(int)));
	connect(this, SIGNAL(newConnectionStatus(int)), infoWidget, SLOT(getNewConnectionStatus(int)));
	connect(this, SIGNAL(newConnection(int)), infoWidget, SLOT(getNewConnection(int)));
	connect(this, SIGNAL(newRemovedConnection(int)), infoWidget, SLOT(getRemovedConnection(int)));
	connect(this, SIGNAL(usedFileStorage(int)), infoWidget, SLOT(getUsedFileStorage(int)));
	connect(this, SIGNAL(usedMemoryStorage(int)), infoWidget, SLOT(getUsedMemoryStorage(int)));
	connect(this, SIGNAL(forwardConnectionList(QList <ConnectionInfo>)), infoWidget, SLOT(getConnectionList(QList <ConnectionInfo>)));
#ifdef PDAGUI
	connect(infoWidget, SIGNAL(sendConnectionList(QList <ConnectionInfo>)), 
			connectionWidget, SLOT(getConnectionList(QList <ConnectionInfo>)));
#endif //PDAGUI
	
	//Bundle List Tag
	bundleWidget = new BundleWidget(this);
	connect(this, SIGNAL(sendBundleList(QList<Bundle>)), bundleWidget, SLOT(getBundleList(QList<Bundle>)));
	connect(bundleWidget, SIGNAL(removeBundle(int)), this, SLOT(removeBundleSlot(int)));
	connect(bundleWidget, SIGNAL(removeAllBundles()), this, SLOT(removeAllBundlesSlot()));

	#ifdef DTN_INTERFACE
	//DTN Interface Tag
	dtnInfo = new QTextEdit;
	dtnInfo->setReadOnly(true);
	dtnEntery = 0;
	#endif /*DNT_INTERFACE*/
	
	//About Tag
	info = new QListWidget;
	//info->setSelectionMode(QAbstractItemView::NoSelection);
	QString itemText;
	itemText.append("http://prophet.snc.sapmi.net");
    QListWidgetItem *newItem = new QListWidgetItem;
    newItem->setText(itemText);
    info->insertItem(0, newItem);
	itemText.clear();
	itemText.append("PRoPHET Version: 2.7.0");
	newItem = new QListWidgetItem;
    newItem->setText(itemText);
    info->insertItem(0, newItem);
	itemText.clear();
	itemText.append("Protocol Authors: Anders Lindgren, Avri Doria");
	newItem = new QListWidgetItem;
    newItem->setText(itemText);
    info->insertItem(0, newItem);
	itemText.clear();
	itemText.append("Code Author: Samo Grasic, Mattias Ek");
	newItem = new QListWidgetItem;
    newItem->setText(itemText);
    info->insertItem(0, newItem);
	itemText.clear();
	itemText.append("Lulea University Of Technology, Center For Distance-Spanning Technology");
	newItem = new QListWidgetItem;
    newItem->setText(itemText);
    info->insertItem(0, newItem);


	//Tab Widget
	tabWidget = new QTabWidget;
	tabWidget->addTab(infoWidget,tr("Node info"));
	#ifdef PDAGUI
	tabWidget->addTab(connectionWidget,tr("Connections"));
	#endif //PDAGUI
	tabWidget->addTab(bundleWidget,tr("Bundle list"));
	#ifdef PDAGUI
	tabWidget->addTab(message, tr("NSIM - Send"));
	tabWidget->addTab(messageRead, tr("NSIM - Read"));
	#else
	tabWidget->addTab(message, tr("NSIM"));
	#endif //PDAGUI
	#ifdef DTN_INTERFACE
	tabWidget->addTab(dtnInfo,tr("DTN Interface"));
	#endif /*DTN_INTERFACE*/
	#ifdef PDAGUI
	tabWidget->addTab(configWidget, tr("Config"));
	#endif //PDAGUI
	tabWidget->addTab(nodeTable, tr("Node list"));
	tabWidget->addTab(info, tr("About"));

	
	//Define final layout
	QVBoxLayout *layout = new QVBoxLayout;
#ifdef PDAGUI
	layout->setMargin(0);
	layout->setSpacing(0);
#endif //PDAGUI
    layout->addWidget(tabWidget);    
    setLayout(layout);

}

void DebugWidget::getBundleList(QList<Bundle> list)
{
	emit sendBundleList(list);
}

void DebugWidget::getNodeList(QList<Node> list)
{
	Node tempNode;
	nodeTable->clear();
    QStringList horizontalNodeLabels;
    horizontalNodeLabels << "ID" << "Name" << "Probability" << "RIB Flag";
	nodeTable->setHorizontalHeaderLabels(horizontalNodeLabels); 
    for (int i = 0; i < list.size(); ++i) 
	{
		tempNode=list.at(i);

		QString text = QString("%1").arg((int)tempNode.nodeId);
		QTableWidgetItem *newItem = new QTableWidgetItem(text);
		nodeTable->setItem(i, 0, newItem);

		text = tempNode.getName();
		newItem = new QTableWidgetItem(text);
		nodeTable->setItem(i, 1, newItem);

		text = QString("%1").arg((float)tempNode.probability);
		newItem = new QTableWidgetItem(text);
		nodeTable->setItem(i, 2, newItem);

		text = QString("%1").arg((int)tempNode.flag);
		newItem = new QTableWidgetItem(text);
		nodeTable->setItem(i, 3, newItem);
		//table->resizeRowToContents(i);
	}
}

void DebugWidget::getConnectionList(QList<ConnectionInfo> list)
{
	emit forwardConnectionList(list);
}

void DebugWidget::getStorageSize(int size)
{

}

void DebugWidget::getLog(QString log)
{
	#ifdef DTN_INTERFACE
	if(dtnEntery>DTN)
	{
		dtnInfo->clear();
		dtnEntery=0;
	}
	dtnInfo->append(log);
	dtnEntery++; 
	#endif /*DTN_INTERFACE*/
}

void DebugWidget::forwardAppBundle(Bundle newBundle)
{
	emit sendAppBundle(newBundle);
}

void DebugWidget::getAppBundle(Bundle newBundle)
{
	message->getBundle(newBundle);
}

void DebugWidget::getTraffic(int t)
{
	emit traffic(t);
}

void DebugWidget::getNewConnection(int c)
{
	emit newConnection(c);
}

void DebugWidget::getNewConnectionStatus(int c)
{
	emit newConnectionStatus(c);
}

void DebugWidget::getRemovedConnection(int c)
{
	emit newRemovedConnection(c);
}

void DebugWidget::getUsedMemoryStorage(int s)
{
	emit usedMemoryStorage(s);
}

void DebugWidget::getUsedFileStorage(int s)
{
	emit usedFileStorage(s);
}

void DebugWidget::removeBundleSlot(int s)
{
	emit removeBundle(s);
}

void DebugWidget::removeAllBundlesSlot()
{
	emit removeAllBundles();
}

