#ifndef HELLO_H
#define HELLO_H

#define SYN 1
#define SYNACK 2
#define ACK 3
#define RSTACK 4



#include <QtCore>

//!Object used during the PRoPHET hello procedure. 
/*!
 * The different functions (HFs) of Hello messages:
 * <PRE>
 * SYN    - Beacon Hello, used for allowing other nodes to see us.
 * SYNACK - Response to a SYN Hello.
 * ACK    - Acknowledgement of the SYNACK.
 * RSTACK - ???
 * </PRE>
 */
class Hello : public QObject
{

Q_OBJECT 
	public:
	int reserved;
	int HFunction;
	int timer;
	int senderId;
	int senderType;
	QString senderName;

	Hello(QObject *parent = 0);
	Hello(const Hello& other);
	Hello& operator=(const Hello& other);


	//! Sets the HF of this Hello.
	/*!
	 * @param function the HF of this Hello.
	 */
	void setFunction(int);

	//! Sets the HF of this Hello.
	/*!
	 * @param function the HF of this Hello.
	 */
	int getFunction(); 

	//! Returns the routable DTN name of the sender.
	/*!
	 * Used for updating routing information and making forwarding 
	 * decisions.
	 * @return the routable DTN name of the sender.
	 */
	QString getSenderName();	

	//! Sets the routable DTN name of the sender.
	/*!
	 * Used for updating routing information and making forwarding 
	 * decisions.
	 * @param name a QString containing the senders name.
	 */	
	void setSenderName(QString);

	//! Sets the Hello timer of this Hello.
	/*!
	 * @param t the interval between periodic Hello messages in units of 
	 * 100 ms.
	 */
	void setTimer(int);

	//! Returns the interval (in 100 ms) this Hello is sent out.
	/*!
	 * @return the interval (in 100 ms) that periodic hellos are sent out.
	 */
	int getTimer(); 
};

#endif
