#include <QtCore>
#include <QtNetwork>
#include <connection.h>
#include <dataPacket.h>
#include <neighbourAwareness.h>
#include <stdio.h>
#include <readFile.h>
#include <main.h>
#include <messageFile.h>
#ifdef Q_WS_X11
#include <sys/types.h>
#include <sys/stat.h> 
#endif
//extern MessageFile *log;

#define VERSION 6
#define HELLO_DATAGRAM "PROPHET"
//#define ALIVE 15
#define TERMINATOR "yxxyyaxayax"
#define TERMINATOR2 "zsdfgfdgfdg"
struct DatagramHeader
{
	qint32 version;
	qint32 lenght;
	qint32 checkSum;
	qint32 nr;
};

ConnectionInfo::ConnectionInfo(QObject *parent)
{
	ReadFile conf;
	alive=conf.getAlive();
	data.clear();
}


ConnectionInfo& ConnectionInfo::operator=(const ConnectionInfo& other)
{
 id = other.id;
 alive = other.alive;
 ip = other.ip;
 data = other.data;
 tcpSocket = other.tcpSocket;
 isClient = other.isClient;
 conptr= other.conptr;

 return *this;
}

ConnectionInfo::ConnectionInfo(const ConnectionInfo& other)
{
 id = other.id;
 alive = other.alive;
 ip = other.ip;
 data = other.data;
 tcpSocket = other.tcpSocket;
 isClient = other.isClient;
 conptr= other.conptr;
}	


NeighAwareServer::NeighAwareServer(QObject* parent) : QTcpServer(parent)
{
}

void NeighAwareServer::incomingConnection(int socketDescriptor)
{ 
	QTcpSocket* newSocket = new QTcpSocket(this);
	newSocket->setSocketDescriptor(socketDescriptor);
	bool permitNewConnection = false;
	emit newConnectionEstablished(newSocket, permitNewConnection);
	if (permitNewConnection == false)
	{
		newSocket->deleteLater();
		newSocket->disconnectFromHost();
	}
}	
	

NeighAware::NeighAware(BundleManager* bundleMng, NodeManager* nodeMng)
{
	nodeManager = nodeMng;
	bundleManager = bundleMng;
	terminator.clear();
	terminator.append(TERMINATOR);
	terminator.append(TERMINATOR2);

	file = new ReadFile;
	myip=file->getNodeIp();
	myip2=file->getNodeIp2();
	broadcast=file->getBroadcast();
	nodeId=file->getNodeId();
	nodeName=file->getNodeName();
	ALIVE=file->getAlive();
	numid = 0;



	tcpServer = new NeighAwareServer(this);
	bool b = connect(tcpServer, SIGNAL(newConnectionEstablished(QTcpSocket*, bool&)), this, SLOT(newConnectionEstablished(QTcpSocket*, bool&)));
	tcpServer->listen(myip,30001);

	udpSocketp = new QUdpSocket(this);
	udpSocketp->bind(30000);
    connect(udpSocketp, SIGNAL(readyRead()),this, SLOT(processPingDatagrams()));
	
	broadcastTimer = new QTimer(this);
	connect(broadcastTimer, SIGNAL(timeout()), this, SLOT(broadcastDatagram()));
	int timer =file->getBroadcastTimer();
	broadcastTimer->start(timer);

	//DEBUG
	log = new MessageFile;
	log->set("connections.log", 1000000, 5);
}

void NeighAware::newConnectionEstablished(QTcpSocket* newSocket, bool& permitNewConnection)
{ 
	QHostAddress ip;
	ip = newSocket->peerAddress();
	QString bla = ip.toString();
	int index=-1;
	for(int i=0;i<connectionlist.size();i++)
	{
		QString bla = ip.toString();
		if(ip == connectionlist.at(i).ip)
			index=i;
	}
	if(index==-1)
	{
		newSocket->connect(newSocket,SIGNAL(readyRead()),this,SLOT(processPendingDatagrams()));
		newSocket->connect(newSocket,SIGNAL(disconnected()),this,SLOT(stateChanged()));
		newSocket->connect(newSocket,SIGNAL(error(QAbstractSocket::SocketError)),this,SLOT(errorDecode(QAbstractSocket::SocketError)));
		//We add a new connection to the list
		ConnectionInfo connectioninfo;
		//Adding
		connectioninfo.conptr = new Connection(numid,file->getNodeId(),file->getNodeName(),nodeManager,bundleManager,this);
		newSocket->connect(connectioninfo.conptr,SIGNAL(sendDataPacket(DataPacket)),this,SLOT(sendDatagram(DataPacket)));
		newSocket->connect(connectioninfo.conptr,SIGNAL(newConnectionStatus(int)),this,SLOT(getNewConnectionStatus(int)));

		///
		connectioninfo.ip=ip;
		connectioninfo.id=numid;
		connectioninfo.alive=ALIVE;
		connectioninfo.tcpSocket = newSocket;
		connectioninfo.isClient = false;


   			if(connectioninfo.tcpSocket->waitForConnected(500))
			{
			    QHostAddress tmp=connectioninfo.ip;
    			    QString logString;
			    logString.append("New connection(S):");	
			    logString.append(tmp.toString());
			    log->addLog(0,logString);
			    connectionlist << connectioninfo;
				emit newConnection(connectioninfo.id);
			}
			else
			{
			    QHostAddress tmp=connectioninfo.ip;
			    QString logString;
			    logString.append("Drop connection(S):");	
			    logString.append(tmp.toString());
			    log->addLog(0,logString);
			    delete connectioninfo.conptr;
			    delete connectioninfo.tcpSocket;
			}

		
		numid++;
		permitNewConnection=true;
		sendDebugConList(connectionlist);
	}
}

void NeighAware::broadcastDatagram()
{  	
	QByteArray datagram=HELLO_DATAGRAM;
	QString debug = broadcast.toString();
 	udpSocketp->writeDatagram(datagram.data(), datagram.size(), broadcast, 30000);
	emit traffic(datagram.size());
	for(int i=0;i<connectionlist.count();i++)
	{
		connectionlist[i].alive--;
		if(connectionlist[i].alive<=0)
		{
			
			connectionlist[i].tcpSocket->disconnectFromHost();
			emit removedConnection(connectionlist[i].id);
			delete connectionlist[i].tcpSocket;
			delete connectionlist[i].conptr;
			//Emiting removed connection to GUI
			emit removedConnection(connectionlist[i].id);
			connectionlist.removeAt(i);	
			//Emiting empty list to GUI
			if(connectionlist.size()==0)
				emit newConnectionStatus(0);
			sendDebugConList(connectionlist);
		}
		sendDebugConList(connectionlist);
	}
	QFile f("connections.txt");
	if (f.open(QIODevice::WriteOnly) == true)
	{
		for(int i=0;i<connectionlist.size();i++)
		{
			f.write(connectionlist[i].ip.toString().toAscii());
			if (connectionlist[i].isClient == true)
				f.write(QString("\tclient\r\n").toAscii());
			else
				f.write(QString("\tserver\r\n").toAscii());
		}
		f.close();
	}

 }

void NeighAware::processPingDatagrams()
{
	while (udpSocketp->hasPendingDatagrams()) 
	{
        QByteArray datagram;
		
		quint16 port;
		QHostAddress ip;
		bool insert=true;
		datagram.resize(udpSocketp->pendingDatagramSize());
        udpSocketp->readDatagram(datagram.data(), datagram.size(), &ip, &port);
		QString bla = ip.toString();
		//Is the node in the list?
		if ((ip==myip) || (ip==myip2))
		{
			insert=false;
		}
		else
		{
			for(int i=0;i<connectionlist.size();i++)
			{
			 
				if(connectionlist.at(i).ip ==ip)
				{ 
					insert=false;
				}
			}
		}
		//If not we insert it.
		if(insert==true)
		{
			QTcpSocket* newSocket = new QTcpSocket(this);
			newSocket->connect(newSocket,SIGNAL(readyRead()),this,SLOT(processPendingDatagrams()));
			newSocket->connect(newSocket,SIGNAL(disconnected()),this,SLOT(stateChanged()));
			newSocket->connect(newSocket,SIGNAL(error(QAbstractSocket::SocketError)),this,SLOT(errorDecode(QAbstractSocket::SocketError)));
			newSocket->connectToHost(ip, 30001);
			//We add a new connection to the list
			ConnectionInfo connectioninfo;
			//Adding
			connectioninfo.conptr = new Connection(numid,file->getNodeId(),file->getNodeName(),nodeManager,bundleManager,this);
			newSocket->connect(connectioninfo.conptr,SIGNAL(sendDataPacket(DataPacket)),this,SLOT(sendDatagram(DataPacket)));
			newSocket->connect(connectioninfo.conptr,SIGNAL(newConnectionStatus(int)),this,SLOT(getNewConnectionStatus(int)));

			///
			connectioninfo.ip=ip;
			connectioninfo.id=numid;
			connectioninfo.alive=ALIVE;
			connectioninfo.tcpSocket = newSocket;
			connectioninfo.isClient = true;

			
    			if(connectioninfo.tcpSocket->waitForConnected(500))
			{
			    QHostAddress tmp=connectioninfo.ip;
    			    QString logString;
			    logString.append("New connection(C):");	
			    logString.append(tmp.toString());
			    log->addLog(0,logString);
				emit newConnection(connectioninfo.id);
			    connectionlist << connectioninfo;
			}
			else
			{
			    QHostAddress tmp=connectioninfo.ip;
			    QString logString;
			    logString.append("Drop connection(C):");	
			    logString.append(tmp.toString());
			    log->addLog(0,logString);
    			    connectioninfo.tcpSocket->disconnectFromHost();
			    delete connectioninfo.conptr;
			    delete connectioninfo.tcpSocket;
			}
			numid++;
			sendDebugConList(connectionlist);
		}
	}
}


void NeighAware::sendDatagram(DataPacket datapacket)
{  	
	QHostAddress ip;
	for(int i=0;i<connectionlist.size();i++)
	{
		if(connectionlist.at(i).id==datapacket.id) 
		{
			ip=connectionlist.at(i).ip;

			//Preparing heder
			QByteArray newDatagram;
			newDatagram.append(datapacket.data);
			//Appending terminator
			newDatagram.append(terminator);
			//Sending datagram 
			connectionlist[i].tcpSocket->write(newDatagram);
			emit traffic(newDatagram.size());
		}
	}


}



void NeighAware::stateChanged()
{
	QString logString;
	logString.append(">>>>>>>>>>>>State:");
	log->addLog(0,logString);

	for(int i=0;i<connectionlist.count();i++)
	{
		if((connectionlist.at(i).tcpSocket->state() != QAbstractSocket::ConnectedState))    
		{
			connectionlist[i].alive=0;

			sendDebugConList(connectionlist);
		}
	}
}

void NeighAware::errorDecode(QAbstractSocket::SocketError error)
{
	int i=error;
	QString logString;
	logString.append(QString("Error %1\n\r").arg((int)error).toAscii());
	log->addLog(0,logString);
	
	for(int i=0;i<connectionlist.count();i++)
	{
		if((connectionlist.at(i).tcpSocket->state() != QAbstractSocket::ConnectedState))    
		{

			connectionlist[i].alive=0;
			sendDebugConList(connectionlist);
		}
	}

			QFile f("error.txt");
		if (f.open(QIODevice::WriteOnly|QIODevice::Append) == true)
		{
			f.write(QString("Error %1\n\r").arg((int)error).toAscii());
			f.close();
		}

}



void NeighAware::processPendingDatagrams()
{
	for(int i=0;i<connectionlist.size();i++)
	{
		QByteArray packetPart;
		packetPart=connectionlist[i].tcpSocket->readAll();
		emit traffic(packetPart.size());
		if(packetPart.size()>0)
		{
			//Updating alive variable
			connectionlist[i].alive=ALIVE;
			//We add datagram to data
			connectionlist[i].data.append(packetPart);
			int idx;
			idx = connectionlist[i].data.indexOf(terminator,0);
			while(idx != -1)
			{
				//We remove the terminator
				QByteArray tempData;
				tempData=connectionlist[i].data.mid(0,idx);
				connectionlist[i].data.remove(0,(idx+terminator.size()));
				DataPacket send;
				send.data.append(tempData);
				send.id=connectionlist[i].id;
				connectionlist[i].conptr->receiveDatagram(send);
				idx = connectionlist[i].data.indexOf(terminator,0);
			}
			
		}
	}
}

void NeighAware::getNewConnectionStatus(int status)
{
	emit newConnectionStatus(status);
}
		
		

		
