#ifndef NEIGHBOURAWARENESS_H
#define NEIGHBOURAWARENESS_H

//#include <iostream.h>
#include <QtCore>
#include <QtNetwork>
#include <connection.h>
#include <dataPacket.h>
#include <neighbourAwareness.h>
#include <bundleManager.h>
#include <nodeManager.h>
#include <messageFile.h>
#include <readFile.h>


/*! \brief This class carries data concerning a certain connection. 
 * Instantiated for each connection.
 */
class ConnectionInfo : public QObject 
{
	Q_OBJECT
	public:
	ConnectionInfo(QObject *parent = 0);
	ConnectionInfo(const ConnectionInfo& other);
	ConnectionInfo& operator=(const ConnectionInfo& other);
	QHostAddress ip;
	QByteArray data;
	QTcpSocket* tcpSocket;
	QTcpSocket* clientTcpSocket;
	Connection *conptr;
	int alive;
	int id;
	bool isClient;
};

class NeighAwareServer : public QTcpServer
{
	Q_OBJECT

public:
	NeighAwareServer(QObject *parent = 0);

protected slots:
	virtual void incomingConnection(int socketDescripton);

signals:
	void newConnectionEstablished(QTcpSocket* newSocket, bool& permitNewConnection);

};


/*! \brief This class detects when two nodes meet to initiate communication. 
 *
 * A ConnectionInfo and the associated Connection object is created for each 
 * encounter.
 */
class NeighAware : public QObject 
{


Q_OBJECT

public:

	NeighAware(BundleManager*, NodeManager*);
	NodeManager *nodeManager;
	BundleManager *bundleManager;

	QList<ConnectionInfo> connectionlist;
   	QByteArray terminator; 	

	QHostAddress myip;
	QHostAddress myip2;
	QHostAddress broadcast;
	int nodeId;
	QString nodeName;
	int ALIVE;
	int numid;
	ReadFile *file;

	NeighAwareServer *tcpServer;
	QUdpSocket *udpSocketp;
	QTimer *broadcastTimer;
	QTimer *testTimer;
	MessageFile* log;

	int logOption;
	
public slots:
	/*!
	 * Expects notification of when new connections are established. It 
	 * sets up the ConnectionInfo and Connection objects, and also connects 
	 * the signals for the new connection. This slot is connected to the 
	 * TcpServer.
	 * @param newSocket the socket of the TCP connection.
	 * @param permitNewConnection this bool will be true if the connection 
	 * was permitted, otherwise false.
	 */
	void newConnectionEstablished(QTcpSocket* newSocket, bool& permitNewConnection);

   	/*!
	 * Writes any socket errors to the file error.txt.
	 */
	void errorDecode(QAbstractSocket::SocketError);
	
	/*!
	 * Removes items with out of date states from the connectionlist.
	 */
	void stateChanged();
	
	/*!
	 * Broadcasts a HELLO_DATAGRAM on the UDP socket 30000.
	 */
	void broadcastDatagram();
	
	/*!
	 * Sends the specified packet on the TCP socket.
	 */
	void sendDatagram(DataPacket datapacket);

	/*!
	 * Receives packets from all TCP connections and dispatches them to 
	 * the appropriate Connection object.
	 */
	void processPendingDatagrams();
	
	/*!
	 * Processes all datagrams received on the UDP socket and creates new 
	 * Connection objects if neccessary.
	 */
	void processPingDatagrams();
	
	/*!
	 * Used just to forward new connection state to the GUI
	 */
	void getNewConnectionStatus(int);
signals:
	/*!
	 * Sends a list of ConnectionInfo objects to the GUI to be 
	 * displayed/updated.
	 * @param a QList of ConnectionInfo objects.
	 */
	void sendDebugConList(QList<ConnectionInfo>);
	/*!
	 * Passes a datagram reveived on the network interface to the TLV. Not used.
	 * @param dataPacket the packet that was received.
	 */
	void passDatagram(DataPacket);
	
	/*!
	 * Signals the size of every sent or received datagram. Connected to 
	 * the GUI. Used for transfer speed calculations.
	 * @param size The size of the received or sent datagram.
	 */
	void traffic(int);

	/*!
	 * Signals that a connection was removed. Connected to GUI.
	 * @param id the id of the connection that was removed.
	 */
	void removedConnection(int);
	
	/*!
	 * Signals that a new connection with another node has been 
	 * established. Connected to GUI.
	 * @param id the id of the new connection.
	 */
	void newConnection(int);

	/*!
	 * Signals new protocol state in one of the connections
	 */
	void newConnectionStatus(int);

};

#endif
