#ifndef NODE_H
#define NODE_H

#include <QtCore>

/*! \brief This class contains the attributes regarding a certain node in the 
 * network.
 */
class Node : public QObject
{

Q_OBJECT 
	public:
	int nodeId;
	float probability;
	int flag;
	QString nodeName;


	Node(QObject *parent = 0);
	Node(const Node& other);
	Node& operator=(const Node& other);


	/*!
	 * Sets the attributes of this node.
	 * @param id the id of this node.
	 * @param pro the delivery predictability to this node.
	 * @param f the (RIB?)flags of this node.
	 * <PRE>
	 * Flag bit 0: Relay Node
	 * Flag bit 1: Custody Node
	 * Flag bit 2: Internet GW Node
	 * Flag bit 3-7: Reserved
	 * </PRE>
	 * @param the name of this node
	 */
	void setContent(int,float,int,QString);

	//! Returns the id if this node
	int getId();

	//! Returns the flags of this node
	int getFlag(); 
	//! Returns the delivery predictability value for this node.
	float getProbability();
	//!Returns the name of this node
	QString getName();
};

#endif
