#include <QtCore>
#include <nodeManager.h>
#include <node.h>
#include <readFile.h>
#include <math.h>
#include <qmutex.h>

//#define AGING_TIMER 900
//#define PENCOUNTER 0.75
//#define BETA 0.25
//#define GAMMA 0.90

NodeManager::NodeManager()//BundleManager* bundleMng)
{
	ReadFile conf;
	BETA=conf.getBeta();
	PENCOUNTER=conf.getPEncounter();
	GAMMA=conf.getGamma();
	AGING_TIMER=conf.getAgingTimer();
	int WRITETIMER =conf.getWriteToFileTimer();
	int fileOption = conf.getUseFileNodes();
	//Create and connect Hello Timer
	//bundleManager = bundleMng;
	agingTimer = new QTimer();
	storagePath=conf.getStoragePath();
	dir.setPath(storagePath);
	//create the dir if it does not exist
	if(!dir.exists())
		dir.mkpath(dir.path());
	option=conf.getUseFileNodes();
        connect(agingTimer, SIGNAL(timeout()), this, SLOT(aging()));
        agingTimer->start(AGING_TIMER);
	if(fileOption==1)
	{
		readFromFile();
		writeToFileTimer = new QTimer();
		connect(writeToFileTimer, SIGNAL(timeout()), this, SLOT(saveToFile()));
		writeToFileTimer->start(WRITETIMER);
		emit sendDebugNodeList(nodeList);
	}
}


void NodeManager::aging()
{
	for (ushort i = 0; i < nodeList.size(); ++i) 
	{
			//AGE PROBABILITY UPDATE
			Node tempNode;
			tempNode=nodeList.at(i);
			tempNode.probability=tempNode.probability*GAMMA;
			nodeList.replace(i,tempNode);

	}
	//DEBUG
	emit sendDebugNodeList(nodeList);
}

void NodeManager::saveToFile()
{
	QFile file(dir.filePath("nodes.txt"));
	if(file.open(QIODevice::WriteOnly |QIODevice::Truncate| QIODevice::Text))
	{
	    QDateTime realTime;
	    realTime=realTime.currentDateTime ();
	    QString time=realTime.toString("MM/dd/yyyy hh:mm:ss");
	    time.append("\n");
	    file.write(time.toAscii ());
	    for (ushort i = 0; i < nodeList.size(); ++i) 
	    {
		QString nodeString;
		nodeString.append(QString("%1").arg((int)nodeList.at(i).nodeId));
		nodeString.append("\t");
		nodeString.append(nodeList.at(i).nodeName);
		nodeString.append("\t");
		//nodeString.append(QString("%1").arg((float)nodeList.at(i).probability));
		//Avoiding double type printout/////////////////
		int prob=nodeList.at(i).probability*100000000;
		QString probString=QString("0.%1").arg(prob);
		nodeString.append(probString);
		////////////////////////////////////////////////
		nodeString.append("\t");
		nodeString.append(QString("%1").arg((int)nodeList.at(i).flag));
		nodeString.append("\n");
		file.write(nodeString.toAscii ());
	    }
	    file.close();
	}
	//DEBUG
	emit sendDebugNodeList(nodeList);
}

void NodeManager::readFromFile()
{
	ReadFile conf;
	QString storagePath=conf.getStoragePath();
	QDir dir;
	dir.setPath(storagePath);
	QFile file(dir.filePath("nodes.txt"));
	if(file.open(QIODevice::ReadOnly | QIODevice::Text))
	{
	    QString firstLine;
	    QString line;
	    //First we read date and time
	    firstLine=file.readLine();
	    //Then we add nodes
	    line=file.readLine();
	    while(line.size()>0)
	    {
    		//Parsing node
		int pos;
		QString temp;
		Node newNode;
		//Node ID
		pos=line.indexOf("\t",0);
		temp=line.mid(0,pos);
		newNode.nodeId=temp.toInt(0,10);
		line.remove(0,pos+1);
		//Node Name
		pos=line.indexOf("\t",0);
		temp=line.mid(0,pos);
		newNode.nodeName=temp;
		line.remove(0,pos+1);
		//Node Probability
		pos=line.indexOf("\t",0);
		temp=line.mid(0,pos);
		newNode.probability=temp.toFloat();
		line.remove(0,pos+1);
		//Node Flag
		pos=line.indexOf("\t",0);
		temp=line.mid(0,pos);
		newNode.flag=temp.toInt(0,10);
		line.remove(0,pos+1);
		//Adding Node
		nodeList.append(newNode);
		line=file.readLine();
	    }
	    //Aging the nodes
	    int option=conf.getAgeFileNodes();
	    if(option==1)
	    {
		QDateTime oldTime;
		firstLine.chop(1);
		oldTime=oldTime.fromString(firstLine,"MM/dd/yyyy hh:mm:ss");
		QDateTime realTime;
		realTime=realTime.currentDateTime ();
		int dif= oldTime.secsTo(realTime);
		int normalAging = conf.getAgingTimer()/1000;
		int aging=dif/normalAging;
		//Aging all nodes
		for (ushort i = 0; i < nodeList.size(); ++i) 
		{
			//AGE PROBABILITY UPDATE
			Node tempNode;
			tempNode=nodeList.at(i);
			tempNode.probability=tempNode.probability*(pow(GAMMA,aging));
			nodeList.replace(i,tempNode);
		}
	    }
	    emit sendDebugNodeList(nodeList);
	    emit updatedNodeList(this);
	    file.close();
	}
}



int NodeManager::getId(QString str)
{
	for (ushort i = 0; i < nodeList.size(); ++i) 
	{
			QString s;
			QString n;
			s=str;
			n=nodeList.at(i).nodeName;
			if(nodeList.at(i).nodeName==str)
				return nodeList.at(i).nodeId;
	}
	return 0;
}


void NodeManager::addDictionary(int ourNodeId,QList<Node> dictionary)
{
	for (ushort i = 0; i < dictionary.size(); ++i) 
	{
		int isInTheList=0;
		for (ushort n = 0; n < nodeList.size(); ++n) 
		{
			if(dictionary.at(i).nodeId==nodeList.at(n).nodeId)
				isInTheList++;
			
		}
		if((isInTheList==0)&&(dictionary.at(i).nodeId!=ourNodeId))
		{
			nodeList.append(dictionary.at(i));
			//!!!!!!!!!!!!!!!!!!!!!!
				//DEBUG
			emit sendDebugNodeList(nodeList);
			emit updatedNodeList(this);
		}
	}
}

void NodeManager::encounterNode(int encounterNodeId,int encounterNodeType,QString encounterNodeName)
{
	int isInTheList=0;
	for (ushort i = 0; i < nodeList.size(); ++i) 
	{
		//ENCOUNTER PROBABILITY UPDATE
		if(nodeList.at(i).nodeId==encounterNodeId)
		{
				Node tempNode;
				tempNode=nodeList.at(i);
				//ENCOUNTING EQUATION
				tempNode.probability=tempNode.probability+((1-tempNode.probability)*PENCOUNTER);
				nodeList.replace(i,tempNode);
				isInTheList++;
		}
	}
	//UNKNOW NEW NODE
	if(isInTheList==0) 
	{
		Node tempNode;
		tempNode.nodeId=encounterNodeId;
		tempNode.nodeName=encounterNodeName;
		tempNode.flag=-1;
		tempNode.probability=PENCOUNTER;
		nodeList.append(tempNode);
		//!!!!!!!!!!!!!!!!!!!!!!
		//DEBUG
		emit sendDebugNodeList(nodeList);
		emit updatedNodeList(this);
	}
}


void NodeManager::addRIB(QList<Node>RIB,int encounterNodeId)//RIB=P_(B,x)
{
	float P_AB=1;
	//First we find P_AB
	for (ushort i = 0; i < nodeList.size(); ++i) 
	{
		if(nodeList.at(i).nodeId==encounterNodeId)
			P_AB = nodeList.at(i).probability;
	}
	//Then we update all transitivites
	for (ushort i = 0; i < nodeList.size(); ++i) 
	{
		//TRANSITIVE PROBABILITY UPDATE
		if(nodeList.at(i).nodeId!=encounterNodeId)
		{
			int nodeCId = nodeList.at(i).nodeId;
			float P_AC = nodeList.at(i).probability;
			float P_BC = 1;
			float newP_AC;
			int flag=0;
			for (ushort n = 0; n < RIB.size(); ++n) 
			{
				if(RIB.at(n).nodeId==nodeCId)
				{
					P_BC = RIB.at(n).probability;
					flag = RIB.at(n).flag;
				}
			}
			newP_AC=P_AC+((1-P_AC) * P_AB * P_BC * BETA);
			//Adding data to list
			Node tempNode;
			tempNode=nodeList.at(i);
			tempNode.probability=newP_AC;
			tempNode.flag=flag;
			nodeList.replace(i,tempNode);
		}

	}
	//DEBUG
	emit sendDebugNodeList(nodeList);
}


void NodeManager::addNode(Node newNode)
{
	nodeList.append(newNode);
	//!!!!!!!!!!!!!!!!!!!!!!!!!!!
	emit updatedNodeList(this);
	//DEBUG
	emit sendDebugNodeList(nodeList);
}

QList<Node> NodeManager::getNodeList()
{	
	return nodeList;
}

//Gives out only the list of nodes with RIB information
QList<Node> NodeManager::getRIBNodeList()
{	
	QList<Node> newNodeList;
	for (ushort i = 0; i < nodeList.size(); ++i) 
	{
		if(nodeList.at(i).flag != -1)
			newNodeList.append(nodeList.at(i));
	
	}
	return newNodeList;
}

