#ifndef NODEMANAGER_H
#define NODEMANAGER_H

#include <QtCore>
#include <node.h>
//#include <messageFile.cpp>
//#include <bundleManager.h>

//! This class contains a list with routing information of all nodes this node is aware of.
class NodeManager : public QObject
{
	Q_OBJECT
public:
	int AGING_TIMER;
	float PENCOUNTER;
	float BETA;
	float GAMMA;
	QTimer *agingTimer;
	QTimer *writeToFileTimer;
	NodeManager();
	QString storagePath;
	QDir dir;
	int option;

	
	
	//! Adds a node to the node list of this NodeManager
	/*! Also emits signals to update the GUI.
	 * @param node the node to be added
	 */
	void addNode(Node);

	//! Looks up the node id of the specified node.
	/*!
	 * @param str the string name of the node.
	 * @return the node id of the specified node.
	 */
	int getId(QString);

	//! Inserts nodes from a dictionary to the node list.
	/*! Only inserts nodes that were not already in the list.
	 * @param the list of nodes to be added to the node list.
	 */
	void addDictionary(int ourNodeId,QList<Node>);

	//! Handles the reception of a new RIB.
	/*! Updates the transitivities in the node list according to the new RIB.
	 * @param RIB the RIB received from the encountered node.
	 * @param encounterNodeId the id if the encountered node.
	 */
	void addRIB(QList<Node>RIB,int encounterNodeId);

	//! Adjusts the delivery predictability for a encountered node.
	/*! If the node was already in the node list the delivery predictability 
	 * is updated, otherwise the new node is added to the node list.
	 * @param encounterNodeId the id of the encountered node
	 * @param encounterNodeType Currently not used
	 * @param encounterNodeName the string name of the encountered node.
	 */
	void encounterNode(int encounterNodeId,int encounterNodeType,QString encounterNodeName);

	//! Returns the node list of this node manager.
	/*! 
	 * @return the nodelist of this node manager.
	 */
	QList<Node> getNodeList();

	//! Returns the list of nodes that have RIB information.
	/*! Iterates through the node list and returns only the nodes that have 
	 * RIB information stored.
	 * @return the nodes that have RIB information.
	 */
	QList<Node> getRIBNodeList();

	//! the list of nodes (should probably be delared private
	QList<Node> nodeList;


signals:
	/*!
	 * Emits the node list. (Used for updating GUI).
	 */
	void sendDebugNodeList(QList<Node>);
	
	/*!
	 * Emits this NodeManager. (Used for GUI???)
	 */
	void updatedNodeList(NodeManager*);
public slots:
	/*!
	 * Performs aging of the probability of all nodes in the node list.
	 */
	void aging();
		
	/*!
	 * Writes the node list to the file nodes.txt.
	 */
	void saveToFile();
		
	/*!
	 * Imports a nodelist from the file nodes.txt and performs 
	 * appropriate aging on the nodes imported.
	 */
	void readFromFile();
};
#endif
